// GenerateBookList.cs
//
// Copyright (C) 2004  Peter Knowles
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.

using System;
using System.Text;
using System.IO;
using System.Collections;

using OpenLibrie.BookList;
using OpenLibrie.LRFLib;
using OpenLibrie.Utils;

namespace OpenLibrie.Tools {

  /// <summary>
  /// Command Line app to generate booklist (index file)
  /// </summary>
  class GenerateBookList {

    /// <summary>
    /// Displays the usage information
    /// </summary>
    private static void ShowUsage(string errMessage) {
      if (errMessage != "") {
	Console.WriteLine("error: {0}", errMessage);
      }
      Console.WriteLine("Usage: booklistgen.exe [options] DIRECTORY\n");
      Console.WriteLine(" -c\tCreate a Librie directory structure under the specified directory.");
      Console.WriteLine(" -f\tForce overwrite of BookList file if it already exists.");
      Console.WriteLine(" -d\tDump XML information and thumbnail images to current directory.");
      Console.WriteLine(" -v\tPrint filenames as they are processed.");
      Console.WriteLine(" -s\tSort the books by title.");
      Console.WriteLine(" -h\tDisplay this help text.\n");

    }

    /// <summary>
    /// Generate a booklist containing all LRF files in a specified directory
    /// </summary>
    public static int Main(string[] Args) {
      string mode = "generate";
      Boolean overwrite = false;
      Boolean verbose = false;
      Boolean dump = false;
      string directory = null;
      Boolean sortByTitle = false;

      #region ParseArguments
      foreach (string arg in Args) {
	if (arg[0] == '-') {
	  switch (arg) {
	  case "-f":
	    overwrite = true;
	    break;
	  case "-c":
	    mode = "init";
	    break;
	  case "-v":
	    verbose = true;
	    break;
	  case "-s":
	    sortByTitle = true;
	    break;
	  case "-d":
	    dump = true;
	    break;
	  case "-h":
	    ShowUsage("");
	    return 0;
	  default:
	    ShowUsage("Unknown option: " + arg);
	    return 1;
	  }
	} else {
	  if (directory == null) {
	    directory = arg;
	  } else {
	    ShowUsage("Too many arguments");
	    return 1;
	  }
	}
      }
      #endregion
      
      if (directory == null) {
	ShowUsage("No root directory specified");
	return 1;
      } else if (! Directory.Exists(directory)) {
	ShowUsage(directory + " does not exist, or is not a directory.");
	return 1;
      } else {
	LibrieFileStructure lfs = new LibrieFileStructure(directory);
	if (mode == "init") {
	  /* Create Directory Structure */
	  try {
	    lfs.CreateDirStructure(verbose);
	  } catch (Exception e) {
	    Console.WriteLine("error creating structure: " + e.Message);
	    return 1;
	  }
	} else if (mode == "generate") {
	  /* Generate BookList file */
	  try {
	    string booklistFilename = lfs.GetIndexFilename();
	    if (File.Exists(booklistFilename) && (! overwrite)) {
	      Console.WriteLine("error: " + booklistFilename + " already exists (Use -f flag to force overwrite)");
	      return 1;
	    } else {
	      BookList bookList = new BookList();
	      ArrayList files = lfs.GetLRFFiles();
	      foreach (string filename in files) {
		if (verbose) {
		  Console.WriteLine("Found: {0}", filename);
		}
		bookList.AddItem(new BookListItem(new LRFFile(filename), lfs.GetRelativeLocation(filename), dump));
	      }
	      bookList.SaveToFile(booklistFilename, sortByTitle, verbose);
	    }
	  } catch (Exception e) {
	    Console.WriteLine("error making index: " + e.Message);
	    return 1;
	  }
	}
	return 0;
      }
    }
  }
  
}
